@@ -1,4 +1,4 @@
-'''Service using multiprocessing to perform CPU-bound operations
+"""Service using multiprocessing to perform CPU-bound operations
 making use of multiple CPUs (i.e. cores) on the machine.
 
 There are 4 levels of constructs.
@@ -73,9 +73,14 @@ The user's main work is implementing the operations in the "workers".
 Another task (of some trial and error) by the user is experimenting with
 CPU allocations among workers to achieve best performance.
 
+Code in the "workers" should raise exceptions as it normally does, without handling them,
+if it considers the situation to be non-recoverable, e.g. input is of wrong type.
+The exceptions will be funneled through the pipelines and raised to the end-user
+with useful traceback info.
+
 Reference: [Service Batching from Scratch, Again](https://zpz.github.io/blog/batched-service-redesign/).
 This article describes roughly version 0.7.2.
-'''
+"""
 from __future__ import annotations
 import asyncio
 import concurrent.futures
@@ -84,6 +89,7 @@ import multiprocessing
 import multiprocessing.queues
 import queue
 import threading
+import traceback
 from abc import ABCMeta, abstractmethod
 from datetime import datetime
 from queue import Empty
@@ -91,29 +97,68 @@ from time import perf_counter, sleep
 from typing import Sequence, Union, Callable, Type, Any
 
 import psutil
+from overrides import final
 
-from .remote_exception import RemoteException, exit_err_msg
-from .util import is_exception, Thread, TimeoutError, ProcessLogger
+from .util import (
+    MP_SPAWN_CTX,
+    SpawnProcess,
+    Thread,
+    rebuild_exception,
+    TimeoutError,
+)
 from ._queues import SingleLane
 
+# This modules uses the 'spawn' method to create processes.
+
 
 # Set level for logs produced by the standard `multiprocessing` module.
 multiprocessing.log_to_stderr(logging.WARNING)
 
 logger = logging.getLogger(__name__)
 
-NOMOREDATA = b'c7160a52-f8ed-40e4-8a38-ec6b84c2cd87'
+NOMOREDATA = b"c7160a52-f8ed-40e4-8a38-ec6b84c2cd87"
+
+
+class _RemoteException_:
+    """
+    This is a helper class to preserve the traceback info of an Exception object
+    during pickling/unpickling. A useful Exception object of the original type
+    will be re-constructed and raised in the "API function", i.e. the touchpoint
+    with the end-user, such as `Server.call` and `Server.async_call`.
+    """
+
+    def __init__(self, exc: BaseException):
+        tb = "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))
+        exc.__traceback__ = None
+        self.exc = exc
+        self.tb = tb
 
 
 class FastQueue(multiprocessing.queues.SimpleQueue):
-    # Check out os.read, os.write, os.close with file-descriptor args.
-    def __init__(self, *, ctx):
+    """
+    This class reduces some overhead in a particular use-case in this module,
+    where one consumer of the queue greedily grabs elements out of the queue
+    towards a batch-size limit.
+
+    It is not "fast" in the general sense, hence the class may not be useful
+    outside of this module.
+
+    Check out os.read, os.write, os.close with file-descriptor args.
+    """
+
+    def __init__(self, *, ctx=None):
+        if ctx is None:
+            ctx = multiprocessing.get_context("spawn")
         super().__init__(ctx=ctx)
         # Replace Lock by RLock to facilitate batching via greedy `get_many`.
         self._rlock = ctx.RLock()
 
 
 class SimpleQueue(queue.SimpleQueue):
+    """
+    Analogous to `FastQueue`.
+    """
+
     def __init__(self):
         super().__init__()
         self._rlock = threading.RLock()
@@ -124,22 +169,29 @@ class Worker(metaclass=ABCMeta):
     # `__init__` and implement `call`.
 
     @classmethod
-    def run(cls, *,
-            q_in: Union[FastQueue, SimpleQueue],
-            q_out: Union[FastQueue, SimpleQueue],
-            **init_kwargs):
+    def run(
+        cls,
+        *,
+        q_in: Union[FastQueue, SimpleQueue],
+        q_out: Union[FastQueue, SimpleQueue],
+        **init_kwargs,
+    ):
         # For a `ProcessWorker`, `q_in` and `q_out` should be `FastQueue`.
         # For a `ThreadWorker`, `q_in` and `q_out` are either `FastQueue` or `SimpleQueue`.
         obj = cls(**init_kwargs)
         q_out.put(obj.name)
-        # This sends a signal to caller indicating completion of init.
+        # This sends a signal to the caller (or "coordinator")
+        # indicating completion of init.
         obj.start(q_in=q_in, q_out=q_out)
 
-    def __init__(self, *,
-                 batch_size: int = None,
-                 batch_wait_time: float = None,
-                 batch_size_log_cadence: int = 1_000_000):
-        '''
+    def __init__(
+        self,
+        *,
+        batch_size: int = None,
+        batch_wait_time: float = None,
+        batch_size_log_cadence: int = 1_000_000,
+    ):
+        """
         `batch_size`: max batch size; see `call`.
 
         `batch_wait_time`: seconds, may be 0; the total duration
@@ -170,7 +222,7 @@ class Worker(metaclass=ABCMeta):
 
         The `__init__` of a subclass may define additional input parameters;
         they can be passed in through `run`.
-        '''
+        """
         if batch_size is None or batch_size == 0:
             batch_size = 0
             if batch_wait_time is None:
@@ -195,40 +247,42 @@ class Worker(metaclass=ABCMeta):
 
     @abstractmethod
     def call(self, x):
-        # If `self.batch_size == 0`, then `x` is a single
-        # element, and this method returns result for `x`.
-        #
-        # If `self.batch_size > 0` (including 1), then
-        # `x` is a list of input data elements, and this
-        # method returns a list of results corresponding
-        # to the elements in `x`.
-        # However, the service (see near the end of `_start_batch`)
-        # will split the resultant list
-        # into single results for individual elements of `x`.
-        # This is *vectorized* computation, or *batching*,
-        # handled by this service pipeline.
-        #
-        # When batching is enabled, the number of
-        # elements in `x` varies between calls depending on the supply
-        # of data. The list `x` does not have a fixed length.
-        #
-        # Be sure to distinguish this from the case where a single
-        # input is naturally a list. In that case, the output of
-        # the current method is the result corresponding to
-        # the single input `x`. The result could be anything---it
-        # may or may not be a list.
-        #
-        # If a subclass fixes `batch_size` in its `__init__` to be
-        # 0 or nonzero, make sure the current method is implemented
-        # accordingly. If `__init__` has no requirement on the value
-        # of `batch_size`, then the current method needs to check
-        # `self.batch_size` and act accordingly, because it is up to
-        # the uer in `Servlet` to specify a zero or nonzero `batch_size`
-        # for this worker.
-        #
-        # In case of exceptions, unless the user has specific things to do,
-        # do not handle them; just let them happen. They will be handled
-        # in `_start_batch` and `_start_single`.
+        """
+        If `self.batch_size == 0`, then `x` is a single
+        element, and this method returns result for `x`.
+
+        If `self.batch_size > 0` (including 1), then
+        `x` is a list of input data elements, and this
+        method returns a list of results corresponding
+        to the elements in `x`.
+        However, the service (see near the end of `_start_batch`)
+        will split the resultant list
+        into single results for individual elements of `x`.
+        This is *vectorized* computation, or *batching*,
+        handled by this service pipeline.
+
+        When batching is enabled, the number of
+        elements in `x` varies between calls depending on the supply
+        of data. The list `x` does not have a fixed length.
+
+        Be sure to distinguish this from the case where a single
+        input is naturally a list. In that case, the output of
+        the current method is the result corresponding to
+        the single input `x`. The result could be anything---it
+        may or may not be a list.
+
+        If a subclass fixes `batch_size` in its `__init__` to be
+        0 or nonzero, make sure the current method is implemented
+        accordingly. If `__init__` has no requirement on the value
+        of `batch_size`, then the current method needs to check
+        `self.batch_size` and act accordingly, because it is up to
+        the uer in `Servlet` to specify a zero or nonzero `batch_size`
+        for this worker.
+
+        In case of exceptions, unless the user has specific things to do,
+        do not handle them; just let them happen. They will be handled
+        in `_start_batch` and `_start_single`.
+        """
         raise NotImplementedError
 
     def start(self, *, q_in, q_out):
@@ -238,9 +292,9 @@ class Worker(metaclass=ABCMeta):
             else:
                 self._start_single(q_in=q_in, q_out=q_out)
         except KeyboardInterrupt:
-            print(self.name, 'stopped by KeyboardInterrupt')
+            print(self.name, "stopped by KeyboardInterrupt")
             # The process will exit. Don't print the usual
-            # exception stuff as that's not need when user
+            # exception stuff as that's not needed when user
             # pressed Ctrl-C.
 
     def _start_single(self, *, q_in, q_out):
@@ -253,7 +307,7 @@ class Worker(metaclass=ABCMeta):
                 break
 
             uid, x = z
-            if is_exception(x):
+            if isinstance(x, (BaseException, _RemoteException_)):
                 q_out.put((uid, x))
                 continue
 
@@ -262,18 +316,22 @@ class Worker(metaclass=ABCMeta):
                     y = self.call([x])[0]
                 else:
                     y = self.call(x)
-            except Exception:
+            except Exception as e:
                 # There are opportunities to print traceback
-                # and details later using the `RemoteException`
-                # object. Be brief on the logging here.
-                y = RemoteException()
+                # and details later. Be brief on the logging here.
+                y = _RemoteException_(e)
 
             q_out.put((uid, y))
 
     def _start_batch(self, *, q_in, q_out):
         def print_batching_info():
-            logger.info('%d batches with sizes %d--%d, mean %.1f',
-                        n_batches, batch_size_min, batch_size_max, batch_size_mean)
+            logger.info(
+                "%d batches with sizes %d--%d, mean %.1f",
+                n_batches,
+                batch_size_min,
+                batch_size_max,
+                batch_size_mean,
+            )
 
         self._batch_buffer = SingleLane(self.batch_size + 10)
         self._batch_get_called = threading.Event()
@@ -301,8 +359,8 @@ class Worker(metaclass=ABCMeta):
 
                 try:
                     results = self.call(batch)
-                except Exception:
-                    err = RemoteException()
+                except Exception as e:
+                    err = _RemoteException_(e)
                     for uid in uids:
                         q_out.put((uid, err))
                 else:
@@ -313,14 +371,16 @@ class Worker(metaclass=ABCMeta):
                     n_batches += 1
                     batch_size_max = max(batch_size_max, n)
                     batch_size_min = min(batch_size_min, n)
-                    batch_size_mean = (batch_size_mean * (n_batches - 1) + n) / n_batches
+                    batch_size_mean = (
+                        batch_size_mean * (n_batches - 1) + n
+                    ) / n_batches
                     if n_batches >= batch_size_log_cadence:
                         print_batching_info()
                         n_batches = 0
         finally:
             if batch_size_log_cadence and n_batches:
                 print_batching_info()
-            collector_thread.join()
+            _ = collector_thread.result()
 
     def _build_input_batches(self, q_in, q_out):
         threading.current_thread().name = f"{self.name}._build_input_batches"
@@ -347,7 +407,7 @@ class Worker(metaclass=ABCMeta):
                             buffer.put(z)
                             return
                         # Now `z` is a tuple like (uid, x).
-                        if is_exception(z[1]):
+                        if isinstance(z[1], (BaseException, _RemoteException_)):
                             q_out.put(z)
                         else:
                             buffer.put(z)
@@ -415,37 +475,29 @@ class Worker(metaclass=ABCMeta):
 
 class ProcessWorker(Worker):
     @classmethod
-    def run(cls, *,
-            log_passer: ProcessLogger,
-            cpus: Sequence[int] = None,
-            **kwargs):
-        '''
+    def run(cls, *, cpus: Sequence[int] = None, **kwargs):
+        """
         This classmethod runs in the worker process to construct
         the worker object and start its processing loop.
 
-        This function is the parameter `target` to `Process`.
+        This function is the parameter `target` to `SpawnProcess`.
         As such, elements in `init_kwargs` go through pickling,
         hence they should consist
         mainly of small, native types such as string, number,small dict.
         Be careful about passing custom class objects in `init_kwargs`.
-        '''
-        try:
-            log_passer.start()
-            if cpus:
-                psutil.Process().cpu_affinity(cpus=cpus)
-            super().run(**kwargs)
-        finally:
-            log_passer.stop()
+        """
+        if cpus:
+            psutil.Process().cpu_affinity(cpus=cpus)
+        super().run(**kwargs)
 
 
 class ThreadWorker(Worker):
-    '''
+    """
     Use this class if the operation is I/O bound (e.g. calling an external service
     to get some info), and computation is very light compared to the I/O part.
     Another use-case of this class is to perform some very simple and quick
     pre-processing or post-processing.
-    '''
-    pass
+    """
 
 
 def make_threadworker(func: Callable[[Any], Any]):
@@ -485,16 +537,18 @@ PassThrough = make_threadworker(lambda x: x)
 
 
 class ProcessServlet:
-    def __init__(self,
-                 worker_cls: Type[ProcessWorker],
-                 *,
-                 cpus: list = None,
-                 name: str = None,
-                 **kwargs):
-        '''
+    def __init__(
+        self,
+        worker_cls: Type[ProcessWorker],
+        *,
+        cpus: list = None,
+        name: str = None,
+        **kwargs,
+    ):
+        """
         `cpus`: specifies how many processes to create and how they are pinned
                 to specific CPUs. The default is a single unpinned process.
-        '''
+        """
         self._worker_cls = worker_cls
         self._name = name or worker_cls.__name__
         self._cpus = cpus
@@ -502,10 +556,7 @@ class ProcessServlet:
         self._workers = []
         self._started = False
 
-    def start(self, q_in, q_out, *, log_passer, ctx):
-        '''
-        `ctx` is a multiprocessing context passed in ultimately from `Server`.
-        '''
+    def start(self, q_in, q_out):
         assert not self._started
 
         # CPU allocation can look as flexible as this:
@@ -525,41 +576,43 @@ class ProcessServlet:
                 sname = self._name
             else:
                 sname = f"{self._name}-{','.join(map(str, cpu))}"
-            logger.info('adding worker <%s> at CPU %s ...', sname, cpu)
-            w = ctx.Process(
-                target=self._worker_cls.run,
-                name=sname,
-                kwargs={
-                    'q_in': q_in,
-                    'q_out': q_out,
-                    'cpus': cpu,
-                    'log_passer': log_passer,
-                    **self._init_kwargs,
-                },
+            logger.info("adding worker <%s> at CPU %s ...", sname, cpu)
+            self._workers.append(
+                SpawnProcess(
+                    target=self._worker_cls.run,
+                    name=sname,
+                    kwargs={
+                        "q_in": q_in,
+                        "q_out": q_out,
+                        "cpus": cpu,
+                        **self._init_kwargs,
+                    },
+                )
             )
-            self._workers.append(w)
-            w.start()
+            self._workers[-1].start()
             name = q_out.get()
-            logger.debug(f"   ... worker <{name}> is ready")
+            logger.debug("   ... worker <%s> is ready", name)
 
-        logger.info(f"servlet {self._name} is ready")
+        logger.info("servlet %s is ready", self._name)
         self._started = True
 
     def stop(self):
         assert self._started
         for w in self._workers:
-            w.join()
+            _ = w.result()
         self._workers = []
         self._started = False
 
 
 class ThreadServlet:
-    def __init__(self,
-                 worker_cls: Type[ThreadWorker],
-                 *,
-                 num_threads: int = None,
-                 name: str = None,
-                 **kwargs):
+    def __init__(
+        self,
+        worker_cls: Type[ThreadWorker],
+        *,
+        num_threads: int = None,
+        name: str = None,
+        **kwargs,
+    ):
         self._worker_cls = worker_cls
         self._name = name or worker_cls.__name__
         self._num_threads = num_threads or 1
@@ -571,61 +624,60 @@ class ThreadServlet:
         assert not self._started
         for ithread in range(self._num_threads):
             sname = f"{self._name}-{ithread}"
-            logger.info('adding worker <%s> in thread ...', sname)
+            logger.info("adding worker <%s> in thread ...", sname)
             w = Thread(
                 target=self._worker_cls.run,
                 name=sname,
                 kwargs={
-                    'q_in': q_in,
-                    'q_out': q_out,
+                    "q_in": q_in,
+                    "q_out": q_out,
                     **self._init_kwargs,
                 },
             )
-            self._workers.append(w)
             w.start()
+            self._workers.append(w)
             name = q_out.get()
-            logger.debug(f"   ... worker <{name}> is ready")
+            logger.debug("   ... worker <%s> is ready", name)
 
-        logger.info(f"servlet {self._name} is ready")
+        logger.info("servlet %s is ready", self._name)
         self._started = True
 
     def stop(self):
         assert self._started
         for w in self._workers:
-            w.join()
+            _ = w.result()
         self._workers = []
         self._started = False
 
 
 class Sequential:
-    '''
+    """
     A sequence of operations performed in order,
-    the previous op's output becoming the subsequent
-    op's input.
-    '''
+    one op's output becoming the next op's input.
+    """
+
     def __init__(self, *servlets: Servlet):
         assert len(servlets) > 0
         self._servlets = servlets
         self._qs = []
         self._started = False
 
-    def start(self, q_in, q_out, *, log_passer, ctx):
+    def start(self, q_in, q_out):
         assert not self._started
         nn = len(self._servlets)
         q1 = q_in
         for i, s in enumerate(self._servlets):
             if i + 1 < nn:
-                if isinstance(s, ThreadServlet) and isinstance(self._servlets[i + 1], ThreadServlet):
+                if isinstance(s, ThreadServlet) and isinstance(
+                    self._servlets[i + 1], ThreadServlet
+                ):
                     q2 = SimpleQueue()
                 else:
-                    q2 = FastQueue(ctx=ctx)
+                    q2 = FastQueue()
                 self._qs.append(q2)
             else:
                 q2 = q_out
-            if isinstance(s, ThreadServlet):
-                s.start(q1, q2)
-            else:
-                s.start(q1, q2, log_passer=log_passer, ctx=ctx)
+            s.start(q1, q2)
             q1 = q2
         self._started = True
 
@@ -645,11 +697,12 @@ class Sequential:
 
 
 class Ensemble:
-    '''
-    A number of operations performed on the same input
-    in parallel, the list of results are gathered and combined
-    to form a final result.
-    '''
+    """
+    A number of operations are performed on the same input in parallel;
+    the list of results, corresponding to the order of the operators,
+    is returned as the result.
+    """
+
     def __init__(self, *servlets: Servlet):
         assert len(servlets) > 1
         self._servlets = servlets
@@ -663,7 +716,7 @@ class Ensemble:
         self._uid_to_results = {}
         self._threads = []
 
-    def start(self, q_in, q_out, *, log_passer, ctx):
+    def start(self, q_in, q_out):
         assert not self._started
         self._reset()
         self._qin = q_in
@@ -674,9 +727,9 @@ class Ensemble:
                 q2 = SimpleQueue()
                 s.start(q1, q2)
             else:
-                q1 = FastQueue(ctx=ctx)
-                q2 = FastQueue(ctx=ctx)
-                s.start(q1, q2, log_passer=log_passer, ctx=ctx)
+                q1 = FastQueue()
+                q2 = FastQueue()
+                s.start(q1, q2)
             self._qins.append(q1)
             self._qouts.append(q2)
         t = Thread(target=self._dequeue)
@@ -702,10 +755,10 @@ class Ensemble:
                 break
 
             uid, x = v
-            if is_exception(x):
+            if isinstance(x, (BaseException, _RemoteException_)):
                 qout.put((uid, x))
                 continue
-            z = {'y': [None] * nn, 'n': 0}
+            z = {"y": [None] * nn, "n": 0}
             catalog[uid] = z
             for q in qins:
                 q.put((uid, x))
@@ -727,12 +780,12 @@ class Ensemble:
                         return
                     uid, y = v
                     z = catalog[uid]
-                    z['y'][idx] = y
-                    z['n'] += 1
-                    if z['n'] == nn:
+                    z["y"][idx] = y
+                    z["n"] += 1
+                    if z["n"] == nn:
                         # All results for this request have been collected.
                         catalog.pop(uid)
-                        y = z['y']
+                        y = z["y"]
                         qout.put((uid, y))
 
     def stop(self):
@@ -740,7 +793,7 @@ class Ensemble:
         for s in self._servlets:
             s.stop()
         for t in self._threads:
-            t.join()
+            _ = t.result()
         self._reset()
         self._started = False
 
@@ -756,36 +809,29 @@ Servlet = Union[ProcessServlet, ThreadServlet, Sequential, Ensemble]
 
 
 class Server:
-    def __init__(self,
-                 servlet: Servlet,
-                 *,
-                 main_cpu: int = 0,
-                 backlog: int = 1024,
-                 ctx=None,
-                 sys_info_log_cadence: int = 1_000_000,
-                 ):
-        '''
+    def __init__(
+        self,
+        servlet: Servlet,
+        *,
+        main_cpu: int = 0,
+        backlog: int = 1024,
+        sys_info_log_cadence: int = 1_000_000,
+    ):
+        """
         `main_cpu`: specifies the cpu for the "main process",
         i.e. the process in which this server objects resides.
 
         `backlog`: max number of requests concurrently in progress within this server,
             all pipes/servlets/stages combined.
 
-        `ctx`: a multiprocessing context. Leave this at default unless you know what you're doing.
-
         `sys_info_log_cadence`: log worker process system info (cpu/memory utilization)
             every this many requests; if `None`, do not log.
-        '''
+        """
         self._servlet = servlet
 
         assert backlog > 0
         self._backlog = backlog
 
-        if ctx is None:
-            self._ctx = multiprocessing.get_context('spawn')
-        else:
-            self._ctx = ctx
-
         if main_cpu is not None:
             # Pin this coordinating thread to the specified CPUs.
             if isinstance(main_cpu, int):
@@ -803,9 +849,6 @@ class Server:
         # should be used with context manager `__enter__`/`__exit__`
         assert not self._started
 
-        self._worker_logger = ProcessLogger(ctx=self.get_mpcontext())
-        self._worker_logger.start()
-
         self._threads = []
 
         self._uid_to_futures = {}
@@ -825,19 +868,13 @@ class Server:
             self._q_in = queue.Queue()
             self._q_out = queue.Queue()
         else:
-            self._q_in = FastQueue(ctx=self.get_mpcontext())
-            self._q_out = FastQueue(ctx=self.get_mpcontext())
-        if isinstance(self._servlet, ThreadServlet):
-            self._servlet.start(self._q_in, self._q_out)
-        else:
-            self._servlet.start(
-                self._q_in, self._q_out,
-                log_passer=self._worker_logger, ctx=self.get_mpcontext())
+            self._q_in = FastQueue()
+            self._q_out = FastQueue()
+        self._servlet.start(self._q_in, self._q_out)
 
         t = Thread(target=self._gather_output)
         t.start()
         self._threads.append(t)
-
         t = Thread(target=self._onboard_input)
         t.start()
         self._threads.append(t)
@@ -845,29 +882,31 @@ class Server:
         self._started = True
         return self
 
-    def __exit__(self, exc_type=None, exc_value=None, exc_traceback=None):
+    # def __exit__(self, exc_type=None, exc_value=None, exc_traceback=None):
+    def __exit__(self, *args, **kwargs):
         assert self._started
-        msg = exit_err_msg(self, exc_type, exc_value, exc_traceback)
-        if msg:
-            logger.error(msg)
+        # msg = exit_err_msg(self, exc_type, exc_value, exc_traceback)
+        # if msg:
+        #     logger.error(msg)
 
         self._input_buffer.put(NOMOREDATA)
 
         self._servlet.stop()
 
         for t in self._threads:
-            t.join()
-
-        self._worker_logger.stop()
+            _ = t.result()
 
         # Reset CPU affinity.
         psutil.Process().cpu_affinity(cpus=[])
         self._started = False
 
-    def get_mpcontext(self):
+    @classmethod
+    @final
+    def get_mp_context(cls):
         # If subclasses need to use additional Queues, Locks, Conditions, etc,
         # they should create them out of this context.
-        return self._ctx
+        # Subclass should not customize this method. Always use spawned processes.
+        return MP_SPAWN_CTX
 
         # TODO: how to track helper processes created by subclasses, so that
         # the sys info log in `_gather_output` can include them?
@@ -886,22 +925,28 @@ class Server:
         fut = self._enqueue(x, timeout)
         return self._wait_for_result(fut)
 
-    def stream(self, data_stream, /, *,
-               return_x: bool = False,
-               return_exceptions: bool = False,
-               timeout=60,
-               ):
-        '''
+    def stream(
+        self,
+        data_stream,
+        /,
+        *,
+        return_x: bool = False,
+        return_exceptions: bool = False,
+        timeout=60,
+    ):
+        """
         The order of elements in the stream is preserved, i.e.,
         elements in the output stream corresponds to elements
         in the input stream in the same order.
-        '''
+        """
 
         # For streaming, "timeout" is usually not a concern.
         # The concern is overall throughput.
 
         def _enqueue(tasks, return_exceptions):
-            threading.current_thread().name = f"{self.__class__.__name__}.stream._enqueue"
+            threading.current_thread().name = (
+                f"{self.__class__.__name__}.stream._enqueue"
+            )
             # Putting input data in the queue does not need concurrency.
             # The speed of sequential push is as fast as it can go.
             _enq = self._enqueue
@@ -920,7 +965,7 @@ class Server:
                             raise
                     tasks.put((x, fut, timedout))
                 # Exceptions in `fut` is covered by `return_exceptions`.
-                # Uncaptured exceptions will propagate and cause the thread to exit in
+                # Uncaught exceptions will propagate and cause the thread to exit in
                 # exception state. This exception is not covered by `return_exceptions`;
                 # it will be detected in the main thread.
             finally:
@@ -952,7 +997,7 @@ class Server:
             else:
                 try:
                     y = _wait(fut)
-                    # May raise RemoteException or TimeoutError.
+                    # May raise TimeoutError or an exception out of _RemoteException_.
                 except Exception as e:
                     if return_exceptions:
                         if return_x:
@@ -981,15 +1026,15 @@ class Server:
 
         # fut = asyncio.Future()
         fut = concurrent.futures.Future()
-        fut.data = {'t0': t0, 'timeout': timeout}
+        fut.data = {"t0": t0, "timeout": timeout}
         uid = id(fut)
         self._uid_to_futures[uid] = fut
         self._input_buffer.put((uid, x))
         return fut
 
     async def _async_wait_for_result(self, fut):
-        t0 = fut.data['t0']
-        t2 = t0 + fut.data['timeout']
+        t0 = fut.data["t0"]
+        t2 = t0 + fut.data["timeout"]
         while not fut.done():
             timenow = perf_counter()
             if timenow > t2:
@@ -997,7 +1042,7 @@ class Server:
                 raise TimeoutError(f"{timenow - t0:.3f} seconds total")
             await asyncio.sleep(min(0.001, t2 - timenow))
         return fut.result()
-        # This could raise RemoteException.
+        # This could raise an exception originating from _RemoteException_.
 
         # TODO: I don't understand why the following (along with
         # corresponding change in `_async_enqueue`) seems to work but
@@ -1007,7 +1052,6 @@ class Server:
         # t2 = t0 + fut.data['timeout']
         # try:
         #     return await asyncio.wait_for(fut, timeout=max(0, t2 - perf_counter()))
-        #     # this may raise RemoteException
         # except asyncio.TimeoutError:
         #     # `wait_for` has already cancelled `fut`.
         #     raise TimeoutError(f"{perf_counter() - t0:.3f} seconds total")
@@ -1026,21 +1070,21 @@ class Server:
             # because the pipe is full and busy.
 
         fut = concurrent.futures.Future()
-        fut.data = {'t0': t0, 'timeout': timeout}
+        fut.data = {"t0": t0, "timeout": timeout}
         uid = id(fut)
         self._uid_to_futures[uid] = fut
         self._input_buffer.put((uid, x))
         return fut
 
     def _wait_for_result(self, fut):
-        t0 = fut.data['t0']
-        t2 = t0 + fut.data['timeout']
+        t0 = fut.data["t0"]
+        t2 = t0 + fut.data["timeout"]
         try:
             return fut.result(timeout=max(0, t2 - perf_counter()))
-            # this may raise RemoteException
-        except concurrent.futures.TimeoutError:
+            # this may raise an exception originating from _RemoteException_
+        except concurrent.futures.TimeoutError as e:
             fut.cancel()
-            raise TimeoutError(f"{perf_counter() - t0:.3f} seconds total")
+            raise TimeoutError(f"{perf_counter() - t0:.3f} seconds total") from e
 
     def _onboard_input(self):
         qin = self._input_buffer
@@ -1062,28 +1106,38 @@ class Server:
             logcounter = 0
 
             def _log_sys_info():
-                pp = [(w.name, psutil.Process(w.pid))
-                      for w in self._servlet._workers
-                      if not isinstance(w, Thread)]
-                msg = [f"  time from           {ts0}",
-                       f"  time to             {datetime.utcnow()}",
-                       f"  items served        {logcounter:_}",
-                       ]
-                attrs = ['memory_full_info',
-                         'cpu_affinity', 'cpu_percent', 'cpu_times',
-                         'num_threads', 'num_fds', 'io_counters',
-                         'status',
-                         ]
+                pp = [
+                    (w.name, psutil.Process(w.pid))
+                    for w in self._servlet._workers
+                    if not isinstance(w, Thread)
+                ]
+                msg = [
+                    f"  time from             {ts0}",
+                    f"  time to               {datetime.utcnow()}",
+                    f"  items served          {logcounter:_}",
+                ]
+                attrs = [
+                    "memory_full_info",
+                    "cpu_affinity",
+                    "cpu_percent",
+                    "cpu_times",
+                    "num_threads",
+                    "num_fds",
+                    "io_counters",
+                    "status",
+                ]
                 for pname, pobj in pp:
-                    msg.append(f'  process {pname}')
+                    msg.append(f"  process {pname}")
                     info = pobj.as_dict(attrs)
                     for k in attrs:
                         v = info[k]
-                        if k == 'memory_full_info':
-                            msg.append(f"        {'memory_uss':<15} {v.uss / 1_000_000:.2f} MB")
+                        if k == "memory_full_info":
+                            msg.append(
+                                f"        {'memory_uss':<15} {v.uss / 1_000_000:.2f} MB"
+                            )
                         else:
                             msg.append(f"        {k:<15} {v}")
-                logger.info('worker process stats:\n%s', '\n'.join(msg))
+                logger.info("worker process stats:\n%s", "\n".join(msg))
 
         try:
             while True:
@@ -1093,12 +1147,21 @@ class Server:
                 uid, y = z
                 fut = futures.pop(uid)
                 try:
-                    if is_exception(y):
+                    if isinstance(y, BaseException):
+                        # Unexpected situation; to be investigated.
+                        logger.warning(
+                            "A non-remote exception has occurred, likely a bug: %r", y
+                        )
                         fut.set_exception(y)
+                    elif isinstance(y, _RemoteException_):
+                        fut.set_exception(rebuild_exception(y.exc, y.tb))
                     else:
                         fut.set_result(y)
-                    fut.data['t1'] = perf_counter()
-                except (concurrent.futures.InvalidStateError, asyncio.InvalidStateError) as e:
+                    fut.data["t1"] = perf_counter()
+                except (
+                    concurrent.futures.InvalidStateError,
+                    asyncio.InvalidStateError,
+                ) as e:
                     if fut.cancelled():
                         # Could have been cancelled due to TimeoutError.
                         # logger.debug('Future object is already cancelled')
@@ -1110,7 +1173,9 @@ class Server:
 
                 if log_cadence:
                     logcounter += 1
-                    if (logcounter >= log_cadence and q_out.empty()) or logcounter >= log_cadence * 10:
+                    if (
+                        logcounter >= log_cadence and q_out.empty()
+                    ) or logcounter >= log_cadence * 10:
                         _log_sys_info()
                         logcounter = 0
                         ts0 = datetime.utcnow()
