@@ -13,14 +13,14 @@ from typing import Iterable, Union, Sequence, Callable, Awaitable, Any
 from overrides import EnforceOverrides
 
 from .util import get_docker_host_ip, is_exception, is_async, MAX_THREADS
-from .remote_exception import RemoteException, exit_err_msg
+from .util import RemoteException
 from ._queues import SingleLane
 
 logger = logging.getLogger(__name__)
 
 
 def put_in_queue(q, x, stop_event, timeout=0.1):
-    '''
+    """
     `q` is either a `threading.Queue` or a `multiprocessing.queues.Queue`,
     but not an `asyncio.Queue`, because the latter does not take
     the `timeout` argument.
@@ -33,7 +33,7 @@ def put_in_queue(q, x, stop_event, timeout=0.1):
     early-stop.
 
     Return `True` if successfully put in queue; `False` if early-stop is detected.
-    '''
+    """
     while True:
         try:
             q.put(x, timeout=timeout)
@@ -61,20 +61,20 @@ def put_in_queue(q, x, stop_event, timeout=0.1):
 
 
 def encode(data, encoder):
-    if encoder == 'pickle':
+    if encoder == "pickle":
         return pickle_dumps(data)
-    if encoder == 'utf8':
-        return data.encode('utf8')
-    assert encoder == 'none'
-    return data   # `data` must be bytes
+    if encoder == "utf8":
+        return data.encode("utf8")
+    assert encoder == "none"
+    return data  # `data` must be bytes
 
 
 def decode(data, encoder):
-    if encoder == 'pickle':
+    if encoder == "pickle":
         return pickle_loads(data)
-    if encoder == 'utf8':
-        return data.decode('utf')
-    assert encoder == 'none'
+    if encoder == "utf8":
+        return data.decode("utf")
+    assert encoder == "none"
     return data  # bytes unchanged
 
 
@@ -90,9 +90,9 @@ def decode(data, encoder):
 # of bytes, which should be decoded by `decode` according to the 'encoder'.
 
 
-async def write_record(writer, request_id, data, *, encoder: str = 'pickle'):
+async def write_record(writer, request_id, data, *, encoder: str = "pickle"):
     data_bytes = encode(data, encoder)
-    writer.write(f'{request_id} {len(data_bytes)} {encoder}\n'.encode())
+    writer.write(f"{request_id} {len(data_bytes)} {encoder}\n".encode())
     writer.write(data_bytes)
     await writer.drain()
     # TODO: add timeout?
@@ -103,7 +103,7 @@ async def read_record(reader, *, timeout=None):
     # This may raise `asyncio.TimeoutError` (nothing to read at the moment)
     # or `asyncio.IncompleteReadError` (connection has been closed by
     # the other side).
-    data = await asyncio.wait_for(reader.readuntil(b'\n'), timeout)
+    data = await asyncio.wait_for(reader.readuntil(b"\n"), timeout)
     request_id, num_bytes, encoder = data[:-1].decode().split()
     data = await reader.readexactly(int(num_bytes))
     return request_id, decode(data, encoder)
@@ -118,26 +118,28 @@ async def read_record(reader, *, timeout=None):
 async def run_tcp_server(conn_handler: Callable, host: str, port: int):
     server = await asyncio.start_server(conn_handler, host, port)
     async with server:
-        addrs = ', '.join(str(sock.getsockname()) for sock in server.sockets)
-        logger.info('serving on %s', addrs)
+        addrs = ", ".join(str(sock.getsockname()) for sock in server.sockets)
+        logger.info("serving on %s", addrs)
         await server.serve_forever()
 
 
 async def run_unix_server(conn_handler: Callable, path: str):
-    '''
+    """
     `path`: a file path accessible by both the server and the client,
         which run on the same machine. This is not a regular file,
         and it (including parent directories) does not need to exist.
         In fact, if the file exists, it will be removed first.
-    '''
+    """
     if os.path.exists(path):
-        assert stat.S_ISSOCK(os.stat(path).st_mode), f"file '{path}' exists but is not a socket file"
+        assert stat.S_ISSOCK(
+            os.stat(path).st_mode
+        ), f"file '{path}' exists but is not a socket file"
     else:
         os.makedirs(os.path.dirname(path), exist_ok=True)
     server = await asyncio.start_unix_server(conn_handler, path)
     async with server:
-        addrs = ', '.join(str(sock.getsockname()) for sock in server.sockets)
-        logger.info('serving on %s', addrs)
+        addrs = ", ".join(str(sock.getsockname()) for sock in server.sockets)
+        logger.info("serving on %s", addrs)
         await server.serve_forever()
 
 
@@ -155,9 +157,9 @@ async def open_tcp_connection(host, port, *, timeout=None):
 
 
 async def open_unix_connection(path: str, *, timeout=None):
-    '''
+    """
     `path`: the same string that has been used by `run_unix_server`.
-    '''
+    """
     t0 = perf_counter()
     while True:
         try:
@@ -171,7 +173,7 @@ async def open_unix_connection(path: str, *, timeout=None):
 
 
 class SocketApplication(EnforceOverrides):
-    '''
+    """
     SocketApplication is designed to to used similar to the "application" in a
     HTTP framework. The main API is to register "endpoint" functions by the method
     `add_route`. This allows to back the socket service by multiple functions
@@ -188,18 +190,24 @@ class SocketApplication(EnforceOverrides):
     This class is the intended interface between a socket server and a particular
     application (functions). Usually, user should not customize the class
     SocketServer.
-    '''
-    def __init__(self, *,
-                 on_startup: Sequence[Callable] = None,
-                 on_shutdown: Sequence[Callable] = None,
-                 ):
+    """
+
+    def __init__(
+        self,
+        *,
+        on_startup: Sequence[Callable] = None,
+        on_shutdown: Sequence[Callable] = None,
+    ):
         self.on_startup = on_startup or []
         self.on_shutdown = on_shutdown or []
         self._routes = {}
 
-    def add_route(self, path: str,
-                  route: Union[Callable[[Any], Awaitable[Any]], Callable[[], Awaitable[Any]]]):
-        '''
+    def add_route(
+        self,
+        path: str,
+        route: Union[Callable[[Any], Awaitable[Any]], Callable[[], Awaitable[Any]]],
+    ):
+        """
         `route` is an *async* function that takes a single positional arg,
         and returns a response (which could be `None` if so desired).
         The response should be serializable by the encoder.
@@ -213,30 +221,31 @@ class SocketApplication(EnforceOverrides):
         way necessary.
 
         There is no GET/POST distinction like in the case of HTTP.
-        '''
+        """
         # TODO: support sync "endpoint" functions, if they are I/O bound.
         self._routes[path] = route
 
     async def handle_request(self, path: str, data: Any = None):
-        '''
+        """
         Dispatch the request to a registered route function.
-        '''
+        """
         if data is None:
             return await self._routes[path]()
         return await self._routes[path](data)
 
 
 class SocketServer(EnforceOverrides):
-    def __init__(self,
-                 app: SocketApplication,
-                 *,
-                 path: str = None,
-                 host: str = None,
-                 port: int = None,
-                 backlog: int = None,
-                 shutdown_path: str = '/shutdown',
-                 ):
-        '''
+    def __init__(
+        self,
+        app: SocketApplication,
+        *,
+        path: str = None,
+        host: str = None,
+        port: int = None,
+        backlog: int = None,
+        shutdown_path: str = "/shutdown",
+    ):
+        """
         `backlog`: max concurrent in-progress requests per connection.
             (Note, the client may open many connections.)
             This "concurrency" is in terms of concurrent calls to
@@ -244,7 +253,7 @@ class SocketServer(EnforceOverrides):
 
         The type of the service, between 'tcp' and 'unix', is determined
         by the parameters `path`, `host`, and `port`. See code for details.
-        '''
+        """
         self.app = app
         if path:
             assert not host
@@ -258,12 +267,12 @@ class SocketServer(EnforceOverrides):
             # it should be '127.0.0.1' (I think but did not verify).
             assert port
             if not host:
-                host = '0.0.0.0'  # in Docker
+                host = "0.0.0.0"  # in Docker
             self._path = None
             self._host = host
             self._port = int(port)
         self._backlog = backlog or 512
-        self._encoder = 'pickle'  # encoder when sending responses
+        self._encoder = "pickle"  # encoder when sending responses
         self._n_connections = 0
         self._shutdown_path = shutdown_path
         self.to_shutdown = False
@@ -277,25 +286,27 @@ class SocketServer(EnforceOverrides):
         return self.__repr__()
 
     async def run(self):
-        '''
+        """
         Start the server and let it stay up until shutdown conditions are met.
-        '''
+        """
         if self._path:
             server_task = asyncio.create_task(
-                run_unix_server(self._handle_connection, self._path))
+                run_unix_server(self._handle_connection, self._path)
+            )
         else:
             server_task = asyncio.create_task(
-                run_tcp_server(self._handle_connection, self._host, self._port))
+                run_tcp_server(self._handle_connection, self._host, self._port)
+            )
         for f in self.app.on_startup:
             if is_async(f):
                 await f()
             else:
                 f()
-        logger.info('server %s is ready', self)
+        logger.info("server %s is ready", self)
         try:
             while True:
                 if self.to_shutdown and not self._n_connections:
-                    raise Exception('shutdown requested')
+                    raise Exception("shutdown requested")
                 await asyncio.sleep(0.1)
         except BaseException as e:
             # This should take care of keyboard interrupt and such.
@@ -308,8 +319,8 @@ class SocketServer(EnforceOverrides):
             server_task.cancel()
             if self._path:
                 os.unlink(self._path)
-            logger.info('server %s is stopped', self)
-            if str(e) != 'shutdown requested':
+            logger.info("server %s is stopped", self)
+            if str(e) != "shutdown requested":
                 raise
 
     async def _handle_connection(self, reader, writer):
@@ -317,11 +328,13 @@ class SocketServer(EnforceOverrides):
         # at the request from a client to the server.
         # This method handles requests in that connection.
         if self._path:
-            addr = writer.get_extra_info('sockname')
+            addr = writer.get_extra_info("sockname")
         else:
-            addr = writer.get_extra_info('peername')
+            addr = writer.get_extra_info("peername")
         self._n_connections += 1
-        logger.info('connection %d is openned from client %r', self._n_connections, addr)
+        logger.info(
+            "connection %d is openned from client %r", self._n_connections, addr
+        )
         reqs = asyncio.Queue(self._backlog)
 
         async def _keep_receiving():
@@ -343,12 +356,12 @@ class SocketServer(EnforceOverrides):
                     self.to_shutdown = True
                     t.set_result(None)
                     break
-                else:
-                    f = self.app.handle_request(path, data)
-                    t = asyncio.create_task(f)
-                    await reqs.put((req_id, t))
-                    # The queue size will restrict how many concurrent calls
-                    # to `handle_request` can be in progress.
+
+                f = self.app.handle_request(path, data)
+                t = asyncio.create_task(f)
+                await reqs.put((req_id, t))
+                # The queue size will restrict how many concurrent calls
+                # to `handle_request` can be in progress.
 
         # `write_record` needs to be called sequentially because it's not atomic;
         # that's why we don't use `add_done_callback` on the Futures to do
@@ -365,12 +378,8 @@ class SocketServer(EnforceOverrides):
                 try:
                     z = await t
                 except Exception as e:
-                    if isinstance(e, RemoteException):
-                        z = e
-                    else:
-                        z = RemoteException(e)
-                enc = 'pickle' if isinstance(z, RemoteException) else self._encoder
-                await write_record(writer, req_id, z, encoder=enc)
+                    z = RemoteException(e)
+                await write_record(writer, req_id, z, encoder=self._encoder)
                 # TODO: what if client has closed the connection?
 
         trec = asyncio.create_task(_keep_receiving())
@@ -384,7 +393,7 @@ class SocketServer(EnforceOverrides):
 
         writer.close()
         await writer.wait_closed()
-        logger.info('connection %d from client %r is closed', self._n_connections, addr)
+        logger.info("connection %d from client %r is closed", self._n_connections, addr)
         self._n_connections -= 1
 
 
@@ -393,24 +402,26 @@ def make_server(app: SocketApplication, **kwargs):
 
 
 def run_app(app, **kwargs):
-    '''
+    """
     End user typicall calls this function to start the server.
     The server will stay up until the client requests its shutdown.
-    '''
+    """
     server = make_server(app, **kwargs)
     asyncio.run(server.run())
 
 
 class SocketClient(EnforceOverrides):
-    def __init__(self, *,
-                 path: str = None,
-                 host: str = None,
-                 port: int = None,
-                 num_connections: int = None,
-                 connection_timeout: int = 60,
-                 backlog: int = 2048,
-                 ):
-        '''
+    def __init__(
+        self,
+        *,
+        path: str = None,
+        host: str = None,
+        port: int = None,
+        num_connections: int = None,
+        connection_timeout: int = 60,
+        backlog: int = 2048,
+    ):
+        """
         `path`, `host`, `port`: either `path` is given (for Unix socket),
             or `port` (plus optionally `host`) is given (for Tcp socket).
             These values should, of course, be consistent with the corresponding
@@ -421,7 +432,7 @@ class SocketClient(EnforceOverrides):
             This is meant for waiting for server to be ready, rather than for
             the action of "connecting" itself (which should be fast).
         `backlog`: size of queue for in-progress requests.
-        '''
+        """
         # Experiments showed `max_connections` can be up to 200.
         # This needs to be improved.
         if path:
@@ -445,7 +456,7 @@ class SocketClient(EnforceOverrides):
         self._num_connections = num_connections or MAX_THREADS
         self._connection_timeout = connection_timeout
         self._backlog = backlog
-        self._encoder = 'pickle'  # encoder when sending requests.
+        self._encoder = "pickle"  # encoder when sending requests.
         self._prepare_shutdown = threading.Event()
         self._to_shutdown = threading.Event()
         self._executor = concurrent.futures.ThreadPoolExecutor()
@@ -469,21 +480,22 @@ class SocketClient(EnforceOverrides):
         n = 0
         while n < self._num_connections:
             z = q.get()
-            if z == 'OK':
+            if z == "OK":
                 n += 1
             if is_exception(z):
                 self._to_shutdown.set()
                 concurrent.futures.wait(self._tasks)
                 self._executor.shutdown()
                 raise z
-        logger.info('client %s is ready', self)
+        logger.info("client %s is ready", self)
         return self
 
-    def __exit__(self, exc_type=None, exc_value=None, exc_traceback=None):
-        msg = exit_err_msg(self, exc_type, exc_value, exc_traceback)
-        if msg:
-            logger.error(msg)
+    # def __exit__(self, exc_type=None, exc_value=None, exc_traceback=None):
+    # msg = exit_err_msg(self, exc_type, exc_value, exc_traceback)
+    # if msg:
+    #     logger.error(msg)
 
+    def __exit__(self, *args, **kwargs):
         self._prepare_shutdown.set()
         t0 = perf_counter()
         while not self._pending_requests.empty():
@@ -500,7 +512,6 @@ class SocketClient(EnforceOverrides):
         self._pending_requests = None
 
     def _open_connections(self, q: queue.Queue):
-
         async def _keep_sending(writer):
             pending = self._pending_requests
             # This queue is populated by `_enqueue`.
@@ -546,16 +557,18 @@ class SocketClient(EnforceOverrides):
             try:
                 if self._path:
                     reader, writer = await open_unix_connection(
-                        self._path, timeout=self._connection_timeout)
+                        self._path, timeout=self._connection_timeout
+                    )
                 else:
                     reader, writer = await open_tcp_connection(
-                        self._host, self._port, timeout=self._connection_timeout)
+                        self._host, self._port, timeout=self._connection_timeout
+                    )
             except Exception as e:
                 q.put(e)
                 return
-            q.put('OK')
-            addr = writer.get_extra_info('peername')
-            logger.info('connection %d to server %r is openned', k + 1, addr)
+            q.put("OK")
+            addr = writer.get_extra_info("peername")
+            logger.info("connection %d to server %r is openned", k + 1, addr)
             tw = asyncio.create_task(_keep_sending(writer))
             tr = asyncio.create_task(_keep_receiving(reader))
             try:
@@ -567,7 +580,7 @@ class SocketClient(EnforceOverrides):
                 await tw
             writer.close()
             await writer.wait_closed()
-            logger.info('connection %d to server %r is closed', k + 1, addr)
+            logger.info("connection %d to server %r is closed", k + 1, addr)
 
         async def _main():
             tasks = [_open_connection(i) for i in range(self._num_connections)]
@@ -580,7 +593,9 @@ class SocketClient(EnforceOverrides):
 
     def _enqueue(self, path: str, data, *, timeout=None):
         if not self._pending_requests:
-            raise Exception("Client is not yet started. Please use it in a context manager")
+            raise Exception(
+                "Client is not yet started. Please use it in a context manager"
+            )
         if self._prepare_shutdown.is_set() or self._to_shutdown.is_set():
             raise Exception("Client is closed")
         fut = concurrent.futures.Future()
@@ -590,8 +605,10 @@ class SocketClient(EnforceOverrides):
         self._pending_requests.put(((path, data), fut), timeout=timeout)
         return fut
 
-    def request(self, path: str, data=None, *, enqueue_timeout=None, response_timeout=None):
-        '''
+    def request(
+        self, path: str, data=None, *, enqueue_timeout=None, response_timeout=None
+    ):
+        """
         This could raise `concurrent.futures.TimeoutError`.
         That means result is not available in the specified time,
         but the request may have well been sent to the server.
@@ -608,24 +625,28 @@ class SocketClient(EnforceOverrides):
         Example:
 
             request('/shutdown', response_timeout=0)
-        '''
+        """
         fut = self._enqueue(path, data, timeout=enqueue_timeout)
         if response_timeout is not None and response_timeout <= 0:
-            return
+            return None
         return fut.result(timeout=response_timeout)
 
-    def stream(self, path: str, data: Iterable, *,
-               return_x: bool = False,
-               return_exceptions: bool = False,
-               enqueue_timeout=60,
-               response_timeout=60,
-               ):
-        '''
+    def stream(
+        self,
+        path: str,
+        data: Iterable,
+        *,
+        return_x: bool = False,
+        return_exceptions: bool = False,
+        enqueue_timeout=60,
+        response_timeout=60,
+    ):
+        """
         If `return_x` is `True`, return a stream of `(x, y)` tuples,
         where `x` is the input data, and `y` is a dict with element 'data'.
         If `return_x` is `False`, return a stream of `y`.
         If `return_exceptions` is `True`, `y` could be an Exception object.
-        '''
+        """
         # Refer to `mpserver.Server.stream` for in-code documentation.
         tasks = SingleLane(self._backlog)
         nomore = object()
